{-# LANGUAGE DefaultSignatures, DeriveGeneric, TypeOperators, FlexibleContexts, FlexibleInstances #-}

import GHC.Generics
import Data.List (intersperse)

data JSON =
  JSNull |  
  JSObject [(String,JSON)] |
  JSArray [JSON] |
  JSValue String

instance Show JSON where 
  show (JSNull) = "null"
  show (JSValue s) = s
  show (JSObject s) = '{' : commaSeparatedfields ++ "}" where 
    commaSeparatedfields = concat $ intersperse ", " fields
    fields = map (\(k,v) -> '"' : k ++ "\": " ++ (show v)) s
  show (JSArray vs) = '[' : commaSeparatedItems ++ "]" where
    commaSeparatedItems = concat $ intersperse ", " items 
    items = map show vs

class ToJSON a where
  toJSON :: a -> JSON
  default toJSON :: (Generic a, ToJSON' (Rep a)) => a -> JSON
  toJSON a = toJSON' (from a)

instance ToJSON a => ToJSON [a] where
  toJSON = JSArray . map toJSON

instance {-# OVERLAPS #-} ToJSON [Char] where
  toJSON s = JSValue ('"' : s ++ "\"")

instance ToJSON Int where
  toJSON i = JSValue (show i)

class ToJSON' f where
  toJSON' :: f p -> JSON

-- Empty Type
instance ToJSON' V1 where 
  toJSON' x = undefined

-- Unit Type
instance ToJSON' U1 where
  toJSON' U1 = JSNull

-- Sum Type
instance (ToJSON' f, ToJSON' g) => ToJSON' (f :+: g) where
  toJSON' x = pullTag (json x)
    where json (L1 x) = toJSON' x
          json (R1 x) = toJSON' x
          pullTag (JSObject [(name, JSObject fields)]) =
            JSObject (("tag",toJSON name):fields)          
            
-- Product Type
instance (ToJSON' f, ToJSON' g) => ToJSON' (f :*: g) where
  toJSON' (a :*: b) = case (toJSON' a, toJSON' b) of
    (JSObject [head], JSObject fields) -> JSObject (head:fields)

-- Type Ref
instance (ToJSON c) => ToJSON' (K1 i c) where
  toJSON' (K1 x) = toJSON x

-- Datatype Metadata (multiple constructors)
instance (ToJSON' f, ToJSON' g, Datatype d) => ToJSON' (M1 D d (f :+: g)) where
  toJSON' (M1 x) = toJSON' x

-- Datatype Metadata (single constructor)
instance (ToJSON' f, Datatype d) => ToJSON' (M1 D d f) where
  toJSON' (M1 x) = dropTag $ toJSON' x
    where dropTag (JSObject [(_,x)]) = x

-- Constructor Metadata
instance (ToJSON' f, Constructor c) => ToJSON' (M1 C c f) where
  toJSON' y@(M1 x) = JSObject [(conName y, toJSON' x)]

-- Selector Metadata
instance (ToJSON' f, Selector c) => ToJSON' (M1 S c f) where
  toJSON' y@(M1 x) = JSObject [(selName y, toJSON' x)]

data Person = Person {
  name :: [String],
  age :: Int 
} deriving Generic

instance ToJSON Person
