import scala.meta._
import scala.annotation.StaticAnnotation
import scala.collection.immutable.Seq

// JSON Conversion
sealed trait ToJSON[T] {
  def apply(in: T): String
}

object ToJSON { 
  def instance[T](f: T => String): ToJSON[T] = new ToJSON[T] {
    def apply(in: T) = f(in)
  }  

  def apply[T](t: T)(implicit toJSON: ToJSON[T]) = toJSON(t)

  implicit val stringToJSON = 
    ToJSON.instance[String]('"' + _ + '"') // Will break with escapes

  implicit val intToJSON = 
    ToJSON.instance[Int](_.toString)

  implicit def listToJSON[T](implicit toJSON: ToJSON[T]) = 
    ToJSON.instance[List[T]](l => l.map(apply(_)).mkString("[",",","]"))

  class Derive extends scala.annotation.StaticAnnotation {
    inline def apply(defn: Any): Any = meta { defn match {      
      case t@q"case class $name(..$params)"=>
        val setters = params.map { param =>
          q""""\""+ ${param.name.syntax} + "\":" + ToJSON(x.${Term.Name(param.name.value)})"""
        }
        val instance = q"""
          implicit def toJSON: ToJSON[$name] = ToJSON.instance(x =>
            Seq("{\n\"constructor\":\"" + ${name.syntax} + "\"",
                ..$setters).mkString(",\n") + "\n}"
          )
        """
        val companion = 
          q"object ${Term.Name(name.value)} { $instance }"          
        Term.Block(Seq(t,companion))
      case _ => abort("Cannot apply @ToJSON here")
    } }
  }
}